# SPDX-FileCopyrightText: 2024-present ZanSara <github@zansara.dev>
# SPDX-License-Identifier: AGPL-3.0-or-later

import pytest
from intentional_core import IntentRouter


@pytest.mark.parametrize(
    "test_configs",
    [
        IntentRouter(
            {
                "stages": {
                    "ask_for_name": {
                        "accessible_from": ["_start_"],
                        "goal": "Ask the user for their name",
                        "outcomes": {
                            "success": {
                                "description": "The user says their name",
                                "move_to": "ask_for_age",
                            },
                            "failure": {
                                "description": "The user states clearly they're not going to share their name",
                                "move_to": "_end_",
                            },
                        },
                    },
                    "ask_for_age": {
                        "goal": "Ask the user for their age",
                        "outcomes": {
                            "success": {
                                "description": "The user says their age",
                                "move_to": "_end_",
                            },
                            "failure": {
                                "description": "The user states clearly they're not going to share their age",
                                "move_to": "_end_",
                            },
                        },
                    },
                }
            }
        )
    ],
)
def test_draw(test_configs):
    pass



# SPDX-FileCopyrightText: 2024-present ZanSara <github@zansara.dev>
# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Sample tools for Intentional's examples.
"""
from typing import Dict, Any, Optional
from datetime import datetime
import structlog
from intentional_core.tools import Tool, ToolParameter


log = structlog.get_logger(logger_name=__name__)


class GetCurrentDateTimeTool(Tool):
    """
    Simple tool to get the current date and time.
    """

    name = "get_current_date_and_time"
    description = "Get the current date and time in the format 'YYYY-MM-DD HH:MM:SS'."
    parameters = []

    async def run(self, params: Optional[Dict[str, Any]] = None) -> str:
        """
        Returns the current time.
        """
        current_datetime = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        log.debug("Getting the current date and time.", current_date_time=current_datetime)
        return current_datetime


class RescheduleInterviewTool(Tool):
    """
    Mock tool to reschedule an interview.
    """

    name = "reschedule_interview"
    description = "Set a new date and time for the interview in the database."
    parameters = [
        ToolParameter(
            "date",
            "The new date for the interview.",
            "string",
            True,
            None,
        ),
        ToolParameter(
            "time",
            "The new time for the interview.",
            "string",
            True,
            None,
        ),
    ]

    async def run(self, params: Optional[Dict[str, Any]] = None) -> str:
        """
        Returns the current time.
        """
        log.debug("Rescheduling the interview.")
        return "The interview was rescheduled successfully."



# SPDX-FileCopyrightText: 2024-present ZanSara <github@zansara.dev>
# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Entry point for the Intentional CLI.
"""
import json
import asyncio
import logging
import logging.config
import argparse

import yaml
import structlog
from intentional_core import load_configuration_file, IntentRouter
from intentional_core.utils import import_plugin

from intentional.draw import to_image


log = structlog.get_logger(logger_name=__name__)


def main():
    """
    Entry point for the Intentional CLI.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("path", help="the path to the configuration file to load.", type=str)
    parser.add_argument("--draw", help="Draw the graph.", action="store_true")
    parser.add_argument("--log-cli-level", help="Select the logging level to the console.", type=str, default="INFO")
    parser.add_argument("--log-file", help="Path to the log file.", type=str)
    parser.add_argument(
        "--log-file-level",
        help="Select the logging level to the file. Ignore if no path is specified with --log-file",
        type=str,
        default="DEBUG",
    )
    args = parser.parse_args()

    # Set the CLI log level
    cli_level = logging.getLevelName(args.log_cli_level.upper())
    structlog.configure(wrapper_class=structlog.make_filtering_bound_logger(cli_level))

    if args.log_file:
        # https://www.structlog.org/en/stable/standard-library.html
        file_level = logging.getLevelName(args.log_file_level.upper())
        timestamper = structlog.processors.TimeStamper(fmt="iso")
        pre_chain = [
            structlog.stdlib.add_log_level,
            structlog.stdlib.ExtraAdder(),
            timestamper,
        ]
        logging.config.dictConfig(
            {
                "version": 1,
                "disable_existing_loggers": False,
                "handlers": {
                    "default": {
                        "level": cli_level,
                        "class": "logging.StreamHandler",
                        "formatter": "colored",
                    },
                    "file": {
                        "level": file_level,
                        "class": "logging.handlers.WatchedFileHandler",
                        "filename": args.log_file,
                        "formatter": "json",
                    },
                },
                "formatters": {
                    "json": {
                        "()": "pythonjsonlogger.jsonlogger.JsonFormatter",
                        "fmt": "%(asctime)s %(levelname)s %(message)s",
                    },
                    "colored": {
                        "()": structlog.stdlib.ProcessorFormatter,
                        "processors": [
                            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                            structlog.dev.ConsoleRenderer(colors=True),
                        ],
                        "foreign_pre_chain": pre_chain,
                    },
                },
                "loggers": {
                    "": {
                        "handlers": ["default", "file"],
                        "level": "DEBUG",
                        "propagate": True,
                    },
                },
            }
        )
        structlog.configure(
            processors=[
                structlog.stdlib.add_log_level,
                structlog.stdlib.PositionalArgumentsFormatter(),
                timestamper,
                structlog.processors.StackInfoRenderer(),
                structlog.processors.format_exc_info,
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
            ],
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
            cache_logger_on_first_use=True,
        )

    if args.draw:
        asyncio.run(draw_intent_graph_from_config(args.path))
        return

    bot = load_configuration_file(args.path)
    asyncio.run(bot.run())


async def draw_intent_graph_from_config(path: str) -> IntentRouter:
    """
    Load the intent router from the configuration file.
    """
    log.debug("Loading YAML configuration file", config_file_path=path)

    with open(path, "r", encoding="utf-8") as file:
        config = yaml.safe_load(file)
    log.debug("Loading bot interface", bot_interface_config=json.dumps(config, indent=4))

    plugins = config.pop("plugins")
    log.debug("Collected_plugins", plugins=plugins)
    for plugin in plugins:
        import_plugin(plugin)

    intent_router = IntentRouter(config.pop("conversation", {}))
    return await to_image(intent_router, path + ".png")



# SPDX-FileCopyrightText: 2024-present ZanSara <github@zansara.dev>
# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Helpers that allow the user to draw the bot's graph.
"""

import base64
from pathlib import Path

import requests
import structlog
from intentional_core.intent_routing import IntentRouter


log = structlog.get_logger(logger_name=__name__)


MERMAID_STYLED_TEMPLATE = """
graph TD;

{connections}

classDef special fill:#aaf,stroke:#aaf;
"""


async def to_image(intent_router: IntentRouter, image_path: Path) -> None:
    """
    Saves an image of the intent's graph at the given path.

    Args:
        intent_router: the intents graph to draw.
        image_path: where to save the resulting image
    """
    image = await to_bytes(intent_router)
    with open(image_path, "wb") as imagefile:
        imagefile.write(image)


async def to_bytes(intent_router: IntentRouter, mermaid_domain: str = "https://mermaid.ink/img/") -> bytes:
    """
    Uses mermaid.ink to render the intent's graph into an image.

    Args:
        intent_router: the intents graph to draw.
        mermaid_domain: the domain of your Mermaid instance, if you have your own.
            Defaults to the public mermaid.ink domain.

    Returns:
        The bytes of the resulting image. To save them into an image file, do:

        ```python
        image = to_image(intent_router)
        with open(image_path, "wb") as imagefile:
            imagefile.write(image)
        ```
    """
    url = to_mermaid_link(intent_router, mermaid_domain)
    resp = requests.get(url, timeout=10)
    if resp.status_code >= 400:
        resp.raise_for_status()
    return resp.content


def to_mermaid_link(intent_router: IntentRouter, mermaid_domain: str = "https://mermaid.ink/img/") -> str:
    """
    Generated a URL that contains a rendering of the graph of the intents into an image.

    Args:
        intent_router: the intents graph to draw.
        mermaid_domain: the domain of your Mermaid instance, if you have your own.
            Defaults to the public mermaid.ink domain.

    Returns:
        A URL on Mermaid.ink with the graph of the intents.
    """
    graph_styled = to_mermaid_diagram(intent_router)
    graphbytes = graph_styled.encode("ascii")
    base64_bytes = base64.b64encode(graphbytes)
    base64_string = base64_bytes.decode("ascii")
    return mermaid_domain + base64_string


def to_mermaid_diagram(intent_router: IntentRouter) -> str:
    """
    Creates a textual representation of the intents graph in a way that Mermaid.ink can render.

    Keep in mind that this function should be able to render also malformed graphs as
    far as possible, because it can be used as a debugging tool to visualize bad bot
    configurations within error messages.

    Args:
        intent_router: the intents graph to draw.

    Returns:
        A string containing the description of the graph.
    """
    stages = {node: f'{node}["<b>{node}</b>"]' for node in intent_router.stages}
    unique_counter = 0
    connections_list = []

    processed_nodes = set()
    for origin, target, key in intent_router.graph.edges:
        # 'end' is reserved in Mermaid
        if target == "end":
            edge_string = f'{stages[origin]} -- {key} --> END["<b>end</b>"]'
        elif target == "_backtrack_":
            unique_counter += 1
            edge_string = f'{stages[origin]} -- {key} --> BACKTRACK{unique_counter}("<b>backtrack</b>"):::special'

        else:
            edge_string = f"{stages[origin]} -- {key} --> {stages.get(target, target)}"
        connections_list.append(edge_string)

        # Indirect connections
        if origin not in processed_nodes:
            processed_nodes.add(origin)

            if intent_router.stages[origin].accessible_from == ["_start_"]:
                edge_string = f'START("<b>start</b>"):::special ---> {stages[origin]}'
            elif intent_router.stages[origin].accessible_from == ["_all_"]:
                unique_counter += 1
                edge_string = f'ALL{unique_counter}("<b>all</b>"):::special ---> {stages[origin]}'
            elif intent_router.stages[origin].accessible_from:
                unique_counter += 1
                accessible_from_str = ",<br>".join(intent_router.stages[origin].accessible_from)
                edge_string = f'FROM{unique_counter}("<b>{accessible_from_str}</b>"):::special ---> {stages[origin]}'
            else:
                continue
            connections_list.append(edge_string)

    connections = "\n".join(connections_list)
    graph_styled = MERMAID_STYLED_TEMPLATE.format(connections=connections)
    log.debug("Mermaid graph created", mermaid_graph=graph_styled)
    return graph_styled



# SPDX-FileCopyrightText: 2024-present ZanSara <github@zansara.dev>
# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Init file for Intentional's sample tools.
"""

from intentional.sample_tools import GetCurrentDateTimeTool, RescheduleInterviewTool


__all__ = ["GetCurrentDateTimeTool", "RescheduleInterviewTool"]
